/*
** Module   : NativeBuffer.java
** Abstract : descriptor for memptr as an argument or return value to a native function 
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20140107 First version.
*/

package com.goldencode.p2j.library;

import java.io.*;
import java.math.*;

import com.goldencode.p2j.util.*;

/**
 * A descriptor for memptr as an argument or return value to a native function.
 */
public class NativeBuffer
extends BaseNativeType
{
   /** The stored value. */
   protected memptr value; 
   
   /**
    * Default constructor.
    */
   public NativeBuffer()
   {
      super();
      type = LibraryManager.TYPE_POINTER;
   }
   
   /**
    * Access the data element.
    *
    * @return   The buffer.
    */
   public memptr getValue()
   {
      return value;
   }
   
   /**
    * Assign the data element.
    *
    * @param    value
    *           The buffer.
    */
   public void setValue(memptr value)
   {
      this.value = value;
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      value = (memptr) in.readObject();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeObject(value);
   }
   
   /**
    * Report the size (in bytes) of the native memory needed to store this instance.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @return   This is always the size of a pointer on the client platform.
    */
   public int size()
   {
      return LibraryManager.ptrSize();
   }
   
   /**
    * Render the data's address into native memory if an explicit address is provided and
    * return the data's address to the caller.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @param    addr
    *           The memory address to which the data should be rendered, if the caller
    *           is supplying a buffer or 0 if the target address should be returned
    *           directly (without further allocation or writing the address to any
    *           buffer).
    *
    * @return   The memory address of the data value or 0 if there is no data to output.
    */
   public long render(long addr)
   {
      long target = 0;
      
      // null, unknown value, uninitialized or 0-size values are written as a null
      // pointer (any failures for these cases have already been screened out, so
      // by the time we get here it is OK to output the 0 (null pointer address)
      if (value != null && !value.isUnknown() && !value.isUninitialized() && 
          value.lengthOf() > 0)
      {
         target = value.getPointerValue().longValue(); 
      }
      
      // - return parms (always scalar) do not call through here
      // - non-return scalar parm types will have addr == 0 and we just need to return the
      //   target address of our memptr
      // - array elements will be called with addr != 0 because the address needs to be
      //   written to memory controlled by the caller
      
      if (addr != 0)
      {
         memptr local = new memptr();
         local.setPointerValue(addr);
         
         if (size() == 4)
         {
            local.setLong(target, 1);
         }
         else
         {
            local.setInt64(target, 1);
         }
      }
      
      return target;
   }
   
   /**
    * Copy back (if necessary) data from the native memory, if given.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @param    addr
    *           The memory address from which the data should be restored, if the caller
    *           is supplying a buffer or 0 if there is no data to copy back.
    */
   public void restore(long addr)
   {
      // only process if the setup code on the server marked this as a case in which to
      // copy back
      if (copyOut)
      {
         // - return parms (always scalar) will have addr != 0 and we must read the pointer
         //   at that address and set it into our instance data
         // - non-return scalar parm types will have addr == 0 and must have had a valid data
         //   address during render() so there is nothing to do here
         //   target address of our memptr
         // - array elements will be called with addr != 0 because the address needs to be
         //   copied back from memory controlled by the caller
      
         if (addr != 0)
         {
            memptr local = new memptr();
            local.setPointerValue(addr);
            
            int64 val = null;
            
            if (size() == 4)
            {
               val = local.getLong(1);
            }
            else
            {
               val = local.getInt64(1);
            }
            
            if (!val.isUnknown())
            {
               long target = val.longValue();
               
               if (value == null)
                  value = new memptr();
               
               value.setPointerValue(target);
            }
         }
      }
   }
}
