/*
** Module   : NativePrimitive.java
** Abstract : descriptor base class for scalar data as an argument or return value to a native
**            function 
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20131212 First version.
*/

package com.goldencode.p2j.library;

import java.io.*;
import com.goldencode.p2j.util.*;

/**
 * A descriptor base class for scalar data as an argument or return value to a native function.
 */
public abstract class NativePrimitive
extends BaseNativeType
{
   /**
    * Default constructor.
    */
   public NativePrimitive()
   {
   }
   
   /**
    * Render the data into native memory, which will be allocated if the address is not
    * given.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @param    addr
    *           The memory address to which the data should be rendered, if the caller
    *           is supplying a buffer or 0 if the buffer should be allocated.
    *
    * @return   The memory address of the rendered data or 0 if there is no data to
    *           output.
    */
   public long render(long addr)
   {
      int sz = size();
      
      memptr local = new memptr();
      
      if (addr == 0)
      {
         // save off for restore use later
         ref = local;
         
         // allocate the block which must be big enough to store the data
         local.setLength(sz);
         
         addr = local.getPointerValue().longValue();
      }
      else
      {
         local.setPointerValue(addr);
      }
      
      if (copyIn)
      {
         writeNative(local);
      }
      
      return addr;
   }
   
   /**
    * Copy back (if necessary) data from the native memory and deallocate the memory
    * if this instance allocated it in the first place.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @param    addr
    *           The memory address from which the data should be restored, if the caller
    *           is supplying a buffer or 0 if the buffer should already exist.
    */
   public void restore(long addr)
   {
      memptr local = ref;
      
      if (local == null && addr != 0)
      {
         local = new memptr();
         local.setPointerValue(addr);
      }
      
      if (copyOut && local != null)
      {
         readNative(local);
      }
      
      // deallocate memory if we allocated it
      if (ref != null)
         ref.setLength(0);
   }
   
   /**
    * Worker method to read the primitive value from the first index position of the given
    * memory buffer.
    *
    * @param    ptr
    *           The memory buffer to reference.
    */
   protected abstract void readNative(memptr ptr);
   
   /**
    * Worker method to write the primitive value to the first index position of the given
    * memory buffer.
    *
    * @param    ptr
    *           The memory buffer to reference.
    */
   protected abstract void writeNative(memptr ptr);
}
